from __future__ import print_function

import argparse
import codecs
import hashlib
import json
import logging
import os
import re
import sys
from typing import (
    Dict,
    Iterator,
    List,
    Match,
    Optional,
    Pattern,
    Set,
    Tuple,
    TYPE_CHECKING,
    Union,
)

from blessings import Terminal
from distutils.version import LooseVersion

from .exceptions import MacroAttributeError, MacroContentError, MacroError
from .types import JirafsMacroAttributes
from . import __version__, constants

if TYPE_CHECKING:
    from .ticketfolder import TicketFolder


logger = logging.getLogger(__name__)


class PluginError(Exception):
    pass


class PluginValidationError(PluginError):
    pass


class PluginOperationError(PluginError):
    pass


class MacroResult(str):
    _generated_filenames: List[str] = []

    def __new__(
        cls, string: str = None, generated_filenames: Optional[List[str]] = None
    ):
        if string is None:
            string = ""
        if generated_filenames is None:
            generated_filenames = []

        self: MacroResult = super().__new__(cls, string)  # type: ignore
        self._generated_filenames = generated_filenames

        return self

    @property
    def generated_filenames(self) -> List[str]:
        return self._generated_filenames


class CommandResult(str):
    _return_code: Optional[int] = None
    terminal: Optional[Terminal] = None
    cursor: int = 0

    def __new__(
        cls, string=None, return_code=None, cursor=0, no_format=False, **kwargs
    ):
        if string is None:
            string = ""
        if string and not string.endswith("\n"):
            string = string + "\n"

        terminal = Terminal()
        if not no_format:
            kwargs["t"] = terminal
            try:
                string = string.format(**kwargs)
            except KeyError:
                logger.warning(
                    "An error was encountered while attempting to format "
                    "string; returning the original string unformatted. "
                    "The caller may want to use the 'no_format' option if "
                    "the outgoing string includes curly braces.",
                )

        self = str.__new__(cls, string)
        self._return_code = return_code
        self.terminal = terminal
        self.cursor = cursor

        return self

    def _echo(self, message: str) -> None:
        print(message, end="")

    def echo(self):
        self._echo(self[self.cursor :])
        self.cursor = len(self)

        return self

    def add_line(self, the_line: str, no_format: bool = False, **kwargs):
        if not the_line.endswith("\n"):
            the_line = the_line + "\n"

        if not no_format:
            kwargs["t"] = self.terminal
            try:
                the_line = the_line.format(**kwargs)
            except KeyError:
                logger.warning(
                    "An error was encountered while attempting to format "
                    "string; returning the original string unformatted. "
                    "The caller may want to use the 'no_format' option if "
                    "the outgoing string includes curly braces.",
                )

        new_result = CommandResult(the_line, no_format=no_format)
        return self + new_result

    def __add__(self, other):
        joined_strings = super(CommandResult, self).__add__(other)

        return_code = None
        if self.return_code is not None:
            return_code = self.return_code
        if isinstance(other, CommandResult) and other.return_code is not None:
            return_code = other.return_code

        return CommandResult(
            joined_strings, return_code=return_code, cursor=self.cursor, no_format=True
        )

    @property
    def return_code(self) -> int:
        return self._return_code or 0

    @return_code.setter
    def return_code(self, value):
        self._return_code = int(value) if value is not None else None


class JirafsPluginBase(object):
    MIN_VERSION = None
    MAX_VERSION = None

    def __init__(self, ticketfolder, entrypoint_name, **kwargs):
        self.ticketfolder: TicketFolder = ticketfolder
        self.entrypoint_name: str = entrypoint_name

    def validate(self, **kwargs) -> bool:
        if not self.MIN_VERSION or not self.MAX_VERSION:
            raise PluginValidationError(
                "Minimum and maximum version numbers not specified."
            )

        min_version = LooseVersion(self.MIN_VERSION)
        max_version = LooseVersion(self.MAX_VERSION)
        curr_version = LooseVersion(__version__)
        if not min_version <= curr_version < max_version:
            raise PluginValidationError(
                "Plugin '%s' is not compatible with version %s of Jirafs; "
                "minimum version: %s; maximum version %s."
                % (
                    self.entrypoint_name,
                    __version__,
                    self.MIN_VERSION,
                    self.MAX_VERSION,
                ),
            )

        return True

    @property
    def metadata_filename(self) -> str:
        return self.ticketfolder.get_metadata_path(
            "plugin_meta", "%s.json" % self.entrypoint_name,
        )

    def get_configuration(self) -> Dict:
        config = self.ticketfolder.get_config()
        if config.has_section(self.entrypoint_name):
            return dict(config.items(self.entrypoint_name))
        return {}

    def _get_metadata(self):
        try:
            with open(self.metadata_filename, "r") as _in:
                return json.loads(_in.read())
        except (IOError, OSError):
            return {}

    def _set_metadata(self, data):
        with open(self.metadata_filename, "w") as out:
            out.write(json.dumps(data, indent=4, sort_keys=True,))

    @property
    def metadata(self):
        if not hasattr(self, "_metadata"):
            self._metadata = self._get_metadata()

        return self._metadata

    def save(self):
        # We'll also be running transformations for items when processing
        # incoming changes from Jira, but since the 'jira' branch is
        # always a fork from 'master', that may cause our later merge
        # of those two branches to create a merge conflict.  To prevent
        # that, let's just not let metadata be saved if we're not on the
        # master copy.
        if not self.ticketfolder.on_master:
            return

        self._set_metadata(self.metadata)


class Plugin(JirafsPluginBase):
    pass


class CommandPlugin(JirafsPluginBase):
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

    def truncate_field_value(self, original_value, length=30):
        if original_value is None:
            original_value = ""
        elif not isinstance(original_value, str):
            original_value = str(original_value)
        value = original_value.strip()
        for newline in ("\n", "\r"):
            if newline in value:
                value = value[0 : value.find(newline)]

        value = value[0:length]

        if value != original_value:
            value = value[0 : length - 1] + "\u2026"

        return value

    def get_description(self):
        try:
            return self.__doc__.strip()
        except AttributeError:
            raise None

    def add_arguments(self, parser):
        pass

    def parse_arguments(self, parser, extra_args):
        return parser.parse_args(extra_args)

    @classmethod
    def get_command_result(cls, result, original=None):
        if not isinstance(result, CommandResult):
            result = CommandResult(result)

        if original is not None:
            result = original + result

        return result

    @classmethod
    def execute_command(cls, extra_args, jira, path, command_name, **ckwargs):
        from .ticketfolder import TicketFolder

        cmd = cls(entrypoint_name=command_name)

        parser = argparse.ArgumentParser(
            prog=os.path.basename(sys.argv[0]) + " " + command_name,
            description=cmd.get_description(),
        )
        if cmd.auto_instantiate_folder():
            parser.add_argument(
                "--no-migrate", dest="migrate", default=True, action="store_false"
            )
        cmd.add_arguments(parser)
        args = cmd.parse_arguments(parser, extra_args)

        folder = None
        folder_plugins = []
        if cmd.auto_instantiate_folder():
            folder = TicketFolder(path, jira, migrate=args.migrate)
            folder_plugins = folder.plugins

        kwargs = {
            "args": args,
            "folder": folder,
            "jira": jira,
            "path": path,
            "parser": parser,
        }
        pre_method = "pre_%s" % command_name
        post_method = "post_%s" % command_name
        for plugin in folder_plugins:
            if not hasattr(plugin, pre_method):
                continue
            method = getattr(plugin, pre_method)
            result = method(**kwargs)
            if result is not None:
                kwargs = result

        cmd.validate(**kwargs)
        result = cls.get_command_result(cmd.handle(**kwargs))

        for plugin in folder_plugins:
            if not hasattr(plugin, post_method):
                continue
            method = getattr(plugin, post_method)
            post_result = method(result)
            if post_result is not None:
                result = cls.get_command_result(post_result, original=result)

        if getattr(cls, "RUN_FOR_SUBTASKS", False):
            for subfolder in folder.subtasks:
                try:
                    cls.execute_command(
                        extra_args, jira, subfolder.path, command_name, **ckwargs
                    )
                except Exception as e:
                    logger.exception(
                        "Exception encountered while running "
                        "'%s' for ticket subfolder '%s': %s"
                        % (command_name, subfolder.ticket_number, e)
                    )

        return result

    def handle(self, *args, **kwargs) -> None:
        self.cmd(*args, **kwargs)

    def cmd(self, *args, **kwargs) -> None:
        # By default, no return value; just execute and move along
        self.main(*args, **kwargs)

    def main(self, *args, **kwargs) -> None:
        raise NotImplementedError()

    def try_subfolders(self) -> bool:
        return getattr(self, "TRY_SUBFOLDERS", False)

    def auto_instantiate_folder(self) -> bool:
        return getattr(self, "AUTOMATICALLY_INSTANTIATE_FOLDER", True,)


class DirectOutputCommandPlugin(CommandPlugin):
    def cmd(self, *args, **kwargs):
        return self.main(*args, **kwargs)


class MacroPlugin(Plugin):
    TAG_NAME = None
    MATCHERS = [
        (
            r"<jirafs:(?P<start>{tag_name}[^>]*)>(?P<content>.*?)"
            r"</jirafs:(?P<end>{tag_name})>"
        ),
        r"<jirafs:(?P<start>{tag_name}[^/]*)/>",
    ]

    def __init__(self, folder, entrypoint_name, *args, **kwargs):
        self.ticketfolder: TicketFolder = folder
        self.entrypoint_name: str = entrypoint_name
        self._args = args
        self._kwargs = kwargs

    @property
    def tag_name(self) -> str:
        assert isinstance(self.TAG_NAME, str)

        return self.TAG_NAME

    def get_matchers(self) -> List[Pattern]:
        return [
            re.compile(rex.format(tag_name=self.tag_name), re.MULTILINE | re.DOTALL,)
            for rex in self.MATCHERS
        ]

    def get_matches(self, content: str) -> Iterator[Match]:
        for matcher in self.get_matchers():
            yield from matcher.finditer(content)

    def get_attributes(self, tag: str) -> JirafsMacroAttributes:
        state_outer = 1
        state_raw_value = 2
        state_squoted_value = 3
        state_dquoted_value = 4
        state_name = 5
        state_name_ended = 6

        state = state_outer

        escapable = (
            "'",
            '"',
        )
        attributes = {}
        key = ""
        value = ""
        value_is_raw = False
        is_escaped = False

        decoder = codecs.getdecoder("unicode_escape")

        def store_value():
            nonlocal value, key, attributes

            if not value_is_raw:
                attributes[key.strip()] = value
            elif value.strip().upper() == "TRUE":
                attributes[key.strip()] = True
            elif value.strip().upper() == "FALSE":
                attributes[key.strip()] = False
            else:
                attributes[key.strip()] = float(value)

            value = ""
            key = ""

        for tag_length, char in enumerate(tag):
            if char.isspace():
                break

        for char in tag[tag_length + 1 :]:
            if state == state_outer:
                if not char.isspace():
                    state = state_name
                    key += char
                    continue
            elif state == state_name:
                if char == "=":
                    state = state_name_ended
                else:
                    key += char
                continue
            elif state == state_name_ended:
                if not char.isspace():
                    if char == '"':
                        state = state_dquoted_value
                        value_is_raw = False
                        continue
                    elif char == "'":
                        state = state_squoted_value
                        value_is_raw = False
                        continue
                    else:
                        state = state_raw_value
                        value_is_raw = True
                        value += char
                        continue
            elif state == state_raw_value:
                if char.isspace():
                    store_value()
                    state = state_outer
                else:
                    value += char
                continue
            elif state == state_dquoted_value:
                if is_escaped:
                    is_escaped = False
                    if char in escapable:
                        value += char
                    else:
                        value += decoder(b"\\%s" % char.encode("utf-8"))[0]
                    continue

                if char == '"':
                    store_value()
                    state = state_outer
                elif char == "\\":
                    is_escaped = True
                else:
                    value += char
                continue
            elif state == state_squoted_value:
                if is_escaped:
                    is_escaped = False
                    if char in escapable:
                        value += char
                    else:
                        value += decoder(b"\\%s" % char.encode("utf-8"))[0]
                    continue

                if char == "'":
                    store_value()
                    state = state_outer
                elif char == "\\":
                    is_escaped = True
                else:
                    value += char
                continue

        if key or value:
            store_value()

        return attributes

    def get_processed_macro_data(
        self, data: str, attrs: JirafsMacroAttributes, config: Dict
    ) -> Union[MacroResult, str]:
        return self.execute_macro(data, attrs, config)

    def process_text_data(self, content: str, path: Optional[str] = None) -> str:
        if path is None:
            path = self.ticketfolder.path

        config = {
            "generated_path": path,
        }

        def run_replacement(match_data):
            data = match_data.groupdict()

            try:
                attrs = self.get_attributes(data.get("start", ""))
            except Exception as e:
                raise MacroAttributeError("Unknown Error") from e

            if data.get("end") and "src" in attrs:
                raise MacroContentError(
                    "Macro cannot use block element form while "
                    "also specifying the 'src' attribute.  'src' is "
                    "used for specifying an external file to use as "
                    "macro content."
                )

            body = data.get("content")
            if "src" in attrs:
                with open(
                    os.path.join(self.ticketfolder.path, attrs["src"],), "r"
                ) as inf:
                    body = inf.read()

            result = self.get_processed_macro_data(body, attrs, config)
            self.save()  # Save metadata changes
            return result

        try:
            content = content
            for matcher in self.get_matchers():
                content = matcher.sub(run_replacement, content)
            return content
        except MacroError:
            raise
        except Exception as e:
            raise MacroContentError(
                "Error encountered while running macro %s: %s" % (self.tag_name, e)
            ) from e

    def process_text_data_reversal(self, data: str) -> str:
        try:
            return self.execute_macro_reversal(data)
        except NotImplementedError:
            return data

    def execute_macro_reversal(self, data: str) -> str:
        raise NotImplementedError()

    def execute_macro(
        self, data: str, attrs: JirafsMacroAttributes, config: Dict
    ) -> Union[MacroResult, str]:
        raise NotImplementedError()

    def cleanup(self) -> None:
        raise NotImplementedError()

    def cleanup_pre_process(self) -> None:
        raise NotImplementedError()

    def cleanup_post_process(self) -> None:
        return self.cleanup()

    def _generate_attrs_string(self, attrs: JirafsMacroAttributes) -> str:
        params = []

        for k, v in sorted(attrs.items(), key=lambda e: e[0]):
            quoted_value = json.dumps(v)
            params.append(f"{k}={quoted_value}")

        if params:
            return " " + " ".join(params)

        return ""

    def generate_tag_from_data_and_attrs(
        self, data: str, attrs: JirafsMacroAttributes
    ) -> str:
        attrs_string = self._generate_attrs_string(attrs)

        if "src" in attrs:
            return f"<jirafs:{self.tag_name}{attrs_string} />"
        else:
            return (
                f"<jirafs:{self.tag_name}{attrs_string}>"
                f"{data}"
                f"</jirafs:{self.tag_name}>"
            )


class AutomaticReversalMacroPlugin(MacroPlugin):
    def should_rerender(self, data: str, cache_entry: Dict, config: Dict,) -> bool:
        if cache_entry:
            return False

        return True

    def _generate_metadata_key(
        self, data_hash: str, attrs: JirafsMacroAttributes, config: Dict
    ) -> str:
        hashable_attrs = json.dumps(attrs, sort_keys=True).encode("utf-8")

        return hashlib.sha256(
            b"".join([data_hash.encode("utf-8"), hashable_attrs])
        ).hexdigest()

    def store_cache_entry(
        self,
        replacement: str,
        filenames: List[str],
        attrs: JirafsMacroAttributes,
        data_hash: str,
        config: Dict,
    ) -> None:
        metadata_key = self._generate_metadata_key(data_hash, attrs, config)

        is_temp = config["generated_path"] != self.ticketfolder.path

        self.metadata.setdefault("reversal_cache", {})[metadata_key] = {
            "filenames": filenames,
            "attrs": attrs,
            "replacement": replacement,
            "is_temp": is_temp,
        }

        stored_in_session = self.metadata.setdefault("stored_in_session", [])
        if metadata_key not in stored_in_session:
            stored_in_session.append(metadata_key)

    def find_cache_entry(
        self, data: str, attrs: JirafsMacroAttributes, data_hash: str, config: Dict
    ) -> Dict:
        generated_path = config["generated_path"]
        existing_files = os.listdir(
            generated_path if generated_path else self.ticketfolder.path
        )

        metadata_key = self._generate_metadata_key(data_hash, attrs, config)

        try:
            entry = self.metadata.get("reversal_cache", {})[metadata_key]

            for filename in entry.get("filenames", []):
                if filename not in existing_files:
                    raise ValueError("Metadata references non-existent file")
            return entry
        except KeyError:
            raise ValueError("Metadata not found")

    def cleanup_pre_process(self) -> None:
        # Clear the 'stored_in_session' list; immediately
        # after this method is executed, we'll re-generate
        # this key, and can use that for determining
        # which keys are still in use
        self.metadata["stored_in_session"] = []
        self.metadata["replacements"] = {}
        self.save()

    def cleanup_post_process(self) -> None:
        cache = self.metadata.get("reversal_cache", {})

        known_keys = set(cache.keys())
        accessed_keys = set(self.metadata.get("stored_in_session", []))
        obsolete_keys = known_keys - accessed_keys

        active_local_files: Set[str] = set()
        active_temp_files: Set[str] = set()
        for key in accessed_keys:
            entry = cache[key]
            target = active_local_files
            if entry["is_temp"]:
                target = active_temp_files
            for filename in entry["filenames"]:
                target.add(filename)

        obsolete_local_files: Set[str] = set()
        obsolete_temp_files: Set[str] = set()
        for key in obsolete_keys:
            entry = cache[key]
            target = obsolete_local_files
            if entry["is_temp"]:
                target = obsolete_temp_files
            for filename in cache[key]["filenames"]:
                target.add(filename)
            del cache[key]
            logger.debug(
                "%s: deleting cache key %s", self.entrypoint_name, key,
            )

        # Delete _both_ obsolete local & temp files from the local
        # directory since a file can be present in both
        local_to_delete = (
            obsolete_local_files | obsolete_temp_files
        ) - active_local_files
        existing_local_files = os.listdir(self.ticketfolder.path)

        for filename in local_to_delete:
            if filename in existing_local_files:
                os.unlink(os.path.join(self.ticketfolder.path, filename))
                logger.debug(
                    "%s: deleting obsolete local file %s",
                    self.entrypoint_name,
                    filename,
                )

        temp_to_delete = obsolete_temp_files - active_temp_files
        temp_path = self.ticketfolder.get_path(constants.TEMP_GENERATED_FILES)
        existing_files = os.listdir(temp_path)

        for filename in temp_to_delete:
            if filename in existing_files:
                os.unlink(os.path.join(temp_path, filename))
                logger.debug(
                    "%s: deleting obsolete temp file %s",
                    self.entrypoint_name,
                    filename,
                )

        self.save()

    def execute_macro_reversal(self, data: str) -> str:
        for replacement, original in self.metadata.get("replacements", {}).items():
            data = data.replace(
                replacement,
                self.generate_tag_from_data_and_attrs(
                    original["data"], original["attrs"],
                ),
            )

        return data

    def get_processed_macro_data(
        self, data: str, attrs: JirafsMacroAttributes, config: Dict
    ) -> Union[MacroResult, str]:
        hashed = hashlib.sha256(data.encode("utf-8")).hexdigest()

        filenames = []
        replacement = ""

        try:
            metadata = self.find_cache_entry(data, attrs, hashed, config)
            replacement = metadata["replacement"]
            filenames = metadata["filenames"]
        except ValueError:
            metadata = {}

        if self.should_rerender(data, metadata, config):
            replacement = self.execute_macro(data, attrs, config)
            if isinstance(replacement, MacroResult):
                filenames = replacement.generated_filenames

        assert replacement

        if replacement:
            self.store_cache_entry(
                replacement, filenames, attrs, hashed, config,
            )
            self.metadata.setdefault("replacements", {})[replacement] = {
                "data": data,
                "attrs": attrs,
            }

        return replacement

    def get_replacement(
        self, data: str, attrs: JirafsMacroAttributes, config: Dict
    ) -> Tuple[str, str]:
        raise NotImplementedError()


class ImageMacroPlugin(AutomaticReversalMacroPlugin):
    def get_extension_and_image_data(
        self, data: str, attrs: JirafsMacroAttributes
    ) -> Tuple[str, bytes]:
        raise NotImplementedError()

    def execute_macro(
        self, data: str, attrs: JirafsMacroAttributes, config: Dict
    ) -> MacroResult:
        generated_path = config["generated_path"]
        hashed = hashlib.sha256(data.encode("utf-8")).hexdigest()

        (extension, image_data) = self.get_extension_and_image_data(data, attrs)

        filename = attrs.get("filename", f"{self.tag_name}.{hashed}.{extension}")

        assert isinstance(filename, str)
        file_path = os.path.join(generated_path, filename,)
        with open(file_path, "wb") as outf:
            outf.write(image_data)

        return MacroResult(
            f'!{filename}|alt="jirafs:{self.tag_name}"!',
            generated_filenames=[filename],
        )



import os
import re
import shutil
import subprocess
import tempfile
from urllib import parse

from jirafs import constants, exceptions, utils
from jirafs.plugin import CommandPlugin
from jirafs.ticketfolder import TicketFolder


class Command(CommandPlugin):
    """ Clone a new ticketfolder for the specified ticket URL"""

    MIN_VERSION = "2.0.0"
    MAX_VERSION = "3.0.0"
    AUTOMATICALLY_INSTANTIATE_FOLDER = False

    TICKET_RE = re.compile(r".*\/browse\/(\w+-\d+)\/?")

    def handle(self, args, jira, path, **kwargs):
        ticket_url = args.ticket_url[0]

        if not os.path.exists(os.path.realpath(ticket_url)):
            ticket_url_parts = parse.urlparse(ticket_url)
            if not ticket_url_parts.netloc:
                default_server = utils.get_default_jira_server()
                ticket_url = parse.urljoin(default_server, "browse/" + ticket_url + "/")

        path = args.path[0] if args.path else None

        return self.cmd(path, ticket_url, jira)

    def clone_from_issue(self, match, ticket_url, path, jira):
        if not path:
            path = match.group(1)
        path = os.path.realpath(path)
        os.mkdir(path)

        try:
            folder = TicketFolder.initialize_ticket_folder(ticket_url, path, jira)

            utils.run_command_method_with_kwargs("pull", folder=folder)
        except Exception:
            shutil.rmtree(path)
            raise

        folder.log(
            "Issue %s cloned successfully to %s", (folder.issue_url, folder.path,)
        )

        return folder

    def clone_from_git_repository(self, url, path, jira):
        temp_dir = tempfile.mkdtemp()

        subprocess.check_call(
            ["git", "clone", url, temp_dir],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        for branch in ["jira", "master"]:
            subprocess.check_call(
                ["git", "checkout", branch],
                cwd=temp_dir,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )

        issue_url_path = os.path.join(temp_dir, constants.METADATA_DIR, "issue_url")
        with open(issue_url_path, "r") as issue_url_file:
            issue_url = issue_url_file.read()

        match = self.TICKET_RE.match(issue_url)
        if not match:
            shutil.rmtree(temp_dir)
            raise exceptions.NotTicketFolderException(
                "The git repository at %s is not a Jirafs backup." % (url,)
            )

        if path:
            path = os.path.realpath(path)
        else:
            # This will create a subdirectory of the cwd named after
            # the ticket's number
            path = os.path.realpath(match.group(1))

        # Move the temporary clone into the proper place
        os.rename(temp_dir, path)

        # Re-clone the shadow repository
        shadow_path = os.path.join(path, constants.METADATA_DIR, "shadow")
        subprocess.check_call(
            ["git", "clone", path, shadow_path],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        # Move the .git directory to where we hide it
        os.rename(
            os.path.join(path, ".git"),
            os.path.join(path, constants.METADATA_DIR, "git"),
        )

        # Reset the shadow's URL to use a relative path
        subprocess.check_call(
            ["git", "remote", "set-url", "origin", "../git"],
            cwd=shadow_path,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        subprocess.check_call(
            ["git", "checkout", "jira"],
            cwd=shadow_path,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        folder = TicketFolder(path, jira,)
        folder.run_git_command(
            "config",
            "--file=%s" % folder.get_metadata_path("git", "config",),
            "core.excludesfile",
            ".jirafs/gitignore",
        )
        folder.log(
            "Cloned Jirafs ticket folder for %s at %s; on hash %s",
            (
                folder.issue_url,
                folder.path,
                folder.run_git_command("rev-parse", "master",),
            ),
        )
        return folder

    def main(self, path, url, jira):
        match = self.TICKET_RE.match(url)
        if match:
            return self.clone_from_issue(match, url, path, jira,)

        # Try checking if it's a git repository, too
        try:
            subprocess.check_call(
                ["git", "ls-remote", url],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            return self.clone_from_git_repository(url, path, jira,)
        except subprocess.CalledProcessError:
            pass

        raise exceptions.JirafsError(
            "'%s' is neither a valid JIRA ticket URL, "
            "nor Jirafs remote backup" % (url)
        )

    def add_arguments(self, parser):
        parser.add_argument("ticket_url", nargs=1, type=str)
        parser.add_argument(
            "path", nargs="*", type=str,
        )



from textwrap import dedent
from unittest import TestCase

from jirafs.jirafieldmanager import JiraFieldManager


class TestJiraFieldManager(TestCase):
    def test_decode(self):
        encoded_values = dedent(
            """
            * summary (summary):
                This is a test summary
            * Longer Message (longer_message):
                This is a much
                longer message that happens
                to contain newlines.
        """
        )

        expected_result = {
            "summary": "This is a test summary",
            "longer_message": (
                "This is a much\n"
                "longer message that happens\n"
                "to contain newlines."
            ),
        }

        actual_result = JiraFieldManager(encoded_values)

        self.assertEqual(expected_result, actual_result)

    def test_decode_with_json_values(self):
        encoded_values = dedent(
            """
            * Dictionary Field (dictionary_field):
                {
                    "name": "Adam Coddington"
                }
            * Integer Field (integer_field):
                10
            * String Field (string_field):
                Hello
            * List Field (list_field):
                [
                    "Alphabet",
                    {
                        "One": "Two"
                    }
                ]
        """
        )

        expected_result = {
            "dictionary_field": {"name": "Adam Coddington"},
            "integer_field": 10,
            "string_field": "Hello",
            "list_field": ["Alphabet", {"One": "Two",}],
        }

        actual_result = JiraFieldManager(encoded_values)

        self.assertEqual(
            expected_result, actual_result,
        )



from mock import patch

from jirafs.plugin import CommandResult

from .base import BaseTestCase


class TestCommandResult(BaseTestCase):
    def test_echo_cursor(self):
        with patch.object(CommandResult, "_echo") as out:
            lines = [
                "Line 1",
                "Line 2",
            ]
            result = CommandResult(lines[0]).echo()
            result = result.add_line(lines[1]).echo()

            self.assertTrue(len(out.call_args_list), 2)

            for idx, line in enumerate(lines):
                args, _ = out.call_args_list[idx]

                self.assertEqual(args[0], line + "\n")



import json
import os
import shutil
import tempfile

import mock
from mock import patch

from jirafs.plugin import MacroPlugin
from jirafs.utils import run_command_method_with_kwargs

from .base import BaseTestCase


class TestPlugins(BaseTestCase):
    def setUp(self):
        self.arbitrary_ticket_number = "ALPHA-123"
        self.root_folder = tempfile.mkdtemp()
        self.mock_jira = mock.MagicMock()
        self.mock_jira.issue.return_value = self.rehydrate_issue("basic.issue.json")
        self.mock_get_jira = lambda _, config=None: self.mock_jira

        with patch(
            "jirafs.ticketfolder.TicketFolder.get_remotely_changed"
        ) as get_remotely_changed:
            get_remotely_changed.return_value = []
            self.ticketfolder = run_command_method_with_kwargs(
                "clone",
                url="http://arbitrary.com/browse/ALPHA-123",
                jira=self.mock_get_jira,
                path=os.path.join(self.root_folder, self.arbitrary_ticket_number,),
            )

    def test_load_plugins(self):
        existing_plugins = {
            "alpha": mock.Mock(MAX_VERSION="10.0", MIN_VERSION="0.1"),
            "beta": mock.Mock(MAX_VERSION="10.0", MIN_VERSION="0.1"),
            "delta": mock.Mock(MAX_VERSION="10.0", MIN_VERSION="0.1"),
        }

        with patch("jirafs.utils.get_installed_plugins") as gip:
            gip.return_value = existing_plugins
            with patch.object(self.ticketfolder, "get_config") as config:
                config.return_value = mock.Mock(
                    has_section=mock.Mock(return_value=True),
                    items=mock.Mock(
                        return_value=[
                            ("alpha", "on"),
                            ("delta", "enabled"),
                            ("gamma", "yes"),
                        ]
                    ),
                )
                results = self.ticketfolder.load_plugins()

        self.assertEqual(2, len(results))
        self.assertTrue(existing_plugins["alpha"].called)
        self.assertFalse(existing_plugins["beta"].called)
        self.assertTrue(existing_plugins["delta"].called)

    def test_macroplugin_block_text_data(self):
        class UppercaseMacroPlugin(MacroPlugin):
            TAG_NAME = "uppercase"

            def execute_macro(self, data, attrs, config):
                return data.upper()

        macro = UppercaseMacroPlugin(self.ticketfolder, "uppercase")

        content = """hello <jirafs:uppercase>you</jirafs:uppercase> there!"""
        expected_result = """hello YOU there!"""
        actual_result = macro.process_text_data(content)

        self.assertEqual(expected_result, actual_result)

    def test_macroplugin_void_text_data(self):
        my_name = "Adam"

        class NameMacroPlugin(MacroPlugin):
            TAG_NAME = "name"

            def execute_macro(self, data, attrs, config):
                return my_name

        macro = NameMacroPlugin(self.ticketfolder, "name")

        content = """hello <jirafs:name/>!"""
        expected_result = """hello Adam!"""
        actual_result = macro.process_text_data(content)

        self.assertEqual(expected_result, actual_result)

    def test_nongreedy_processing(self):
        class UppercaseMacroPlugin(MacroPlugin):
            TAG_NAME = "uppercase"

            def execute_macro(self, data, attrs, config):
                return data.upper()

        macro = UppercaseMacroPlugin(self.ticketfolder, "uppercase")

        content = """
            hello <jirafs:uppercase>you</jirafs:uppercase>; how are
            you <jirafs:uppercase>doing
            </jirafs:uppercase>?
        """
        expected_result = """
            hello YOU; how are
            you DOING
            ?
        """
        actual_result = macro.process_text_data(content)

        self.assertEqual(expected_result, actual_result)

    def test_attribute_extraction_block(self):
        class TestMacroPlugin(MacroPlugin):
            TAG_NAME = "test"

            def execute_macro(self, data, attrs, config):
                return json.dumps(attrs)

        macro = TestMacroPlugin(self.ticketfolder, "test")

        content = """<jirafs:test alpha="AL''\\"\\tPHA" beta=2 gamma=True epsilon='bloop"\\''>
            beep
            </jirafs:test>
        """

        expected_result = {
            "alpha": "AL''\"\tPHA",
            "beta": 2.0,
            "gamma": True,
            "epsilon": "bloop\"'",
        }
        self.assertEqual(expected_result, json.loads(macro.process_text_data(content)))

    def test_attribute_extraction_void(self):
        class TestMacroPlugin(MacroPlugin):
            TAG_NAME = "test"

            def execute_macro(self, data, attrs, config):
                return json.dumps(attrs)

        macro = TestMacroPlugin(self.ticketfolder, "test")

        content = """
            <jirafs:test alpha="AL''\\"\\tPHA" beta=2 gamma=True epsilon='bloop"\\''/>
        """

        expected_result = {
            "alpha": "AL''\"\tPHA",
            "beta": 2.0,
            "gamma": True,
            "epsilon": "bloop\"'",
        }
        self.assertEqual(expected_result, json.loads(macro.process_text_data(content)))

    def test_multiline_attribute_extraction(self):
        class TestMacroPlugin(MacroPlugin):
            TAG_NAME = "test"

            def execute_macro(self, data, attrs, config):
                return json.dumps(attrs)

        macro = TestMacroPlugin(self.ticketfolder, "test")

        content = """
            <jirafs:test
                alpha="AL''\\"PHA"
                beta=2
                gamma=True
                epsilon='bloop"\\''
            />
        """

        expected_result = {
            "alpha": "AL''\"PHA",
            "beta": 2.0,
            "gamma": True,
            "epsilon": "bloop\"'",
        }
        self.assertEqual(expected_result, json.loads(macro.process_text_data(content)))

    def test_macroplugin_block_text_data_double(self):
        class CounterMacroPlugin(MacroPlugin):
            TAG_NAME = "counter"

            def execute_macro(self, data, attrs, config):
                return str(int(attrs["counter"])) + ":" + data

        macro = CounterMacroPlugin(self.ticketfolder, "counter")

        content = """
        Start
        <jirafs:counter counter=0>Zero</jirafs:counter>
        Middle
        <jirafs:counter counter=1>One</jirafs:counter>
        End
        """
        expected_result = """
        Start
        0:Zero
        Middle
        1:One
        End
        """
        actual_result = macro.process_text_data(content)

        self.assertEqual(expected_result, actual_result)

    def tearDown(self):
        shutil.rmtree(self.root_folder)



import pydoc

from jirafs.plugin import CommandPlugin


class Command(CommandPlugin):
    """ Print the log for this issue """

    MIN_VERSION = "2.0.0"
    MAX_VERSION = "3.0.0"

    def main(self, folder, **kwargs):
        results = folder.get_log()
        pydoc.pager(results)
        return results



from jirafs import utils
from jirafs.plugin import CommandPlugin
from jirafs.utils import run_command_method_with_kwargs


class Command(CommandPlugin):
    """Create a new JIRA issue"""

    MIN_VERSION = "2.0.0"
    MAX_VERSION = "3.0.0"
    AUTOMATICALLY_INSTANTIATE_FOLDER = False

    FIELDS = (
        {
            "name": "project",
            "prompt": "Project",
            "required": True,
            "path": "project.key",
        },
        {
            "name": "issuetype",
            "prompt": "Issue Type",
            "default": "Task",
            "path": "issuetype.name",
        },
        {"name": "summary", "prompt": "Summary", "required": True},
        {"name": "description", "prompt": "Description", "required": False},
    )

    def set_field_value(self, data, field, value):
        starting_reference = data

        path = field.get("path", field["name"])
        for part in path.split("."):
            if part not in data:
                data[part] = {}
            last_reference = data
            data = data[part]

        last_reference[path.split(".")[-1]] = value

        return starting_reference

    def prompt_for_input(self, field):
        while True:
            if field.get("default"):
                value = input(
                    "%s (%s): " % (field.get("prompt"), field.get("default"),)
                )
            else:
                value = input("%s: " % (field.get("prompt")))

            value = value.strip()

            if value:
                return value
            elif not field.get("required") and field.get("default"):
                return field.get("default")
            elif not field.get("required"):
                return value

    def main(self, args, jira, path, parser, **kwargs):
        server = args.server
        if not server:
            server = utils.get_default_jira_server()

        issue_data = {}

        for field in self.FIELDS:
            if getattr(args, field["name"]) is not None:
                self.set_field_value(issue_data, field, getattr(args, field["name"]))
            elif args.quiet:
                self.set_field_value(issue_data, field, field.get("default"))
            else:
                self.set_field_value(issue_data, field, self.prompt_for_input(field))

        jira_client = jira(server)
        issue = jira_client.create_issue(issue_data)

        return run_command_method_with_kwargs(
            "clone", path=None, url=issue.permalink(), jira=jira,
        )

    def add_arguments(self, parser):
        parser.add_argument("--server", default=None)
        parser.add_argument("--quiet", "-q", default=False, action="store_true")
        for argument in self.FIELDS:
            parser.add_argument("--%s" % argument["name"], default=None, type=str)



from __future__ import print_function

import collections
import json
import os
import shutil
import subprocess

from behave import *
from jira.client import JIRA


@given("jirafs is installed and configured")
def installed_and_configured(context):
    pass


@given("a cloned ticket with the following fields")
def cloned_ticket_with_following_fields(context):
    jira_client = JIRA(
        {
            "server": context.integration_testing["url"],
            "verify": False,
            "check_update": False,
        },
        basic_auth=(
            context.integration_testing["username"],
            context.integration_testing["password"],
        ),
    )

    issue_data = {
        "project": {"key": context.integration_testing["project"]},
        "issuetype": {"name": "Task",},
    }
    for row in context.table:
        issue_data[row[0]] = json.loads(row[1])

    issue = jira_client.create_issue(issue_data)

    if not "cleanup_steps" in context:
        context.cleanup_steps = []
    context.cleanup_steps.append(lambda context: issue.delete())
    context.execute_steps(
        u"""
        when the command "jirafs clone {url}" is executed
        and we enter the ticket folder for "{url}"
        """.format(
            url=issue.permalink()
        )
    )


@when('the command "{command}" is executed')
def execute_command(context, command):
    command = command.format(**context.integration_testing)

    env = os.environ.copy()
    env["JIRAFS_GLOBAL_CONFIG"] = context.integration_testing["config_path"]
    env["JIRAFS_ALLOW_USER_INPUT__BOOL"] = "0"

    proc = subprocess.Popen(
        command.encode("utf-8"),
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        env=env,
    )

    stdout, stderr = proc.communicate()

    if not hasattr(context, "executions"):
        context.executions = collections.deque()

    context.executions.appendleft(
        {
            "command": command,
            "stdout": stdout.decode("utf-8"),
            "stderr": stderr.decode("utf-8"),
            "return_code": proc.returncode,
        }
    )


@when('we enter the ticket folder for "{url}"')
def execute_command(context, url):
    url = url.format(**context.integration_testing)

    os.chdir(os.path.join(os.getcwd(), url.split("/")[-1]))


@then('the directory will contain a file named "{filename}"')
def directory_contains_file(context, filename):
    assert filename in os.listdir("."), "%s not in folder" % filename


@then('the output will contain the text "{expected}"')
def output_will_contain(context, expected):
    expected = expected.format(**context.integration_testing)

    assert expected in context.executions[0]["stdout"], "%s not in %s" % (
        expected,
        context.executions[0]["stdout"],
    )


@step("print execution results")
def print_stdout(context):
    print(json.dumps(context.executions[0], indent=4, sort_keys=True))


@step("debugger")
def debugger(context):
    import ipdb

    ipdb.set_trace()



import codecs
import fnmatch
import logging
import logging.handlers
import io
import json
import os
import re
import subprocess
from urllib import parse

from jira.resources import Issue

from . import constants
from . import exceptions
from . import migrations
from . import utils
from .jiralinkmanager import JiraLinkManager
from .jirafieldmanager import JiraFieldManager
from .plugin import MacroPlugin, PluginValidationError
from .exceptions import MacroError


class TicketFolderLoggerAdapter(logging.LoggerAdapter):
    def process(self, msg, kwargs):
        return (
            "{{{issue_id}}} {msg}".format(issue_id=self.extra["issue_id"], msg=msg,),
            kwargs,
        )


class TicketFolder(object):
    def __init__(self, path, jira, migrate=True, quiet=False):
        self.path = os.path.realpath(os.path.expanduser(path))
        self.quiet = quiet
        self.issue_url = self.get_ticket_url()
        self.get_jira = jira

        self._formatter = logging.Formatter(
            fmt="%(asctime)s\t%(levelname)s\t%(module)s\t%(message)s"
        )
        self._handler = logging.handlers.RotatingFileHandler(
            self.get_metadata_path(constants.TICKET_OPERATION_LOG),
            maxBytes=2 ** 20,
            backupCount=2,
        )
        self._handler.setFormatter(self._formatter)
        self._logger = logging.getLogger(
            ".".join([__name__, self.ticket_number.replace("-", "_")])
        )
        self._logger.addHandler(self._handler)
        self._logger_adapter = TicketFolderLoggerAdapter(
            self._logger, {"issue_id": self.ticket_number},
        )

        if not os.path.isdir(self.metadata_dir):
            raise exceptions.NotTicketFolderException(
                "%s is not a synchronizable ticket folder" % (path)
            )

        self.plugins = self.load_plugins()

        if migrate:
            self.run_migrations()

        # If no `new_comment.jira.txt` file exists, let's create one
        comment_path = self.get_local_path(constants.TICKET_NEW_COMMENT)
        if not os.path.exists(comment_path):
            with io.open(comment_path, "w", encoding="utf-8") as out:
                out.write("")

        # Let's update the ignore file while we're here.
        self.build_ignore_files()

    @property
    def logger(self):
        return self._logger_adapter

    def __repr__(self):
        value = self.__str__()
        return "<%s>" % value

    def __str__(self):
        return "[%s] at %s" % (self.ticket_number, self.path)

    @property
    def subtasks(self):
        if hasattr(self, "_subtasks"):
            return self._subtasks
        self._subtasks = []

        subtasks_path = self.get_metadata_path("subtasks")
        if not os.path.exists(subtasks_path):
            return self._subtasks

        with open(subtasks_path, "r") as in_:
            for line in in_:
                ticket_number = line.strip()
                folder = self.__class__(
                    self.get_path(ticket_number,), utils.lazy_get_jira()
                )
                self._subtasks.append(folder)

        return self._subtasks

    def load_plugins(self):
        config = self.get_config()
        plugins = []

        if not config.has_section(constants.CONFIG_PLUGINS):
            return plugins

        installed_plugins = utils.get_installed_plugins()

        for name, status in config.items(constants.CONFIG_PLUGINS):
            if not utils.convert_to_boolean(status):
                # This plugin is not turned on.
                continue
            if name not in installed_plugins:
                # This plugin is not installed.
                self.log(
                    "Plugin '%s' is not available.", (name,),
                )
                continue

            plugin = installed_plugins[name](self, name)

            try:
                plugin.validate()
            except PluginValidationError as e:
                self.log(
                    "Plugin '%s' did not pass validation; not loading: %s.", (name, e,)
                )

            plugins.append(plugin)

        return plugins

    def get_config(self):
        local_config_file = self.get_metadata_path("config")
        additional_configs = []
        if os.path.exists(local_config_file):
            additional_configs.append(local_config_file)

        return utils.get_config(additional_configs)

    def set_config_value(self, section, key, value):
        with utils.stash_local_changes(self):
            local_config_file = self.get_metadata_path("config")

            config = utils.get_config(
                additional_configs=[local_config_file], include_global=False,
            )
            if not config.has_section(section):
                config.add_section(section)
            config.set(section, key, value)

            with open(local_config_file, "w") as out:
                config.write(out)

            self.run_git_command("add", ".jirafs/config")
            self.run_git_command("commit", "-m", "Config change", failure_ok=True)

    @property
    def jira_base(self):
        match = re.match(r"(.*)\/browse\/.*", self.issue_url)
        if not match:
            raise ValueError(
                "Could not infer JIRA server URL from issue URL %s" % (self.issue_url,)
            )
        return match.group(1)

    @property
    def ticket_number(self):
        parts = parse.urlparse(self.issue_url)
        match = re.match(r".*\/browse\/(\w+-\d+)\/?.*", parts.path)
        if not match:
            raise ValueError(
                "Could not infer ticket number from URL %s" % self.issue_url
            )
        return match.group(1)

    @property
    def jira(self):
        if not hasattr(self, "_jira"):
            self._jira = self.get_jira(self.jira_base, config=self.get_config())
        return self._jira

    @property
    def issue(self):
        if not hasattr(self, "_issue"):
            self._issue = self.jira.issue(self.ticket_number)
        return self._issue

    def clear_cache(self):
        if hasattr(self, "_issue"):
            del self._issue
        if hasattr(self, "_jira"):
            del self._jira

    def store_cached_issue(self, shadow=True):
        storable = {"options": self.issue._options, "raw": self.issue.raw}
        with io.open(
            self.get_path(".jirafs/issue.json", shadow=shadow), "w", encoding="utf-8",
        ) as out:
            out.write(
                json.dumps(storable, indent=4, sort_keys=True, ensure_ascii=False,)
            )

    @property
    def cached_issue(self):
        if not hasattr(self, "_cached_issue"):
            try:
                issue_path = self.get_metadata_path("issue.json")
                with io.open(issue_path, "r", encoding="utf-8") as _in:
                    storable = json.loads(_in.read())
                    self._cached_issue = Issue(
                        storable["options"], None, storable["raw"],
                    )
            except IOError:
                self.log(
                    "Error encountered while loading cached issue!",
                    level=logging.ERROR,
                )
                self._cached_issue = self.issue
        return self._cached_issue

    @property
    def metadata_dir(self) -> str:
        return os.path.join(self.path, constants.METADATA_DIR,)

    @property
    def git_master(self) -> str:
        return self.run_git_command("rev-parse", "master")

    @property
    def git_merge_base(self) -> str:
        return self.run_git_command("merge-base", "master", "jira",)

    @property
    def git_branch(self) -> str:
        return self.run_git_command("rev-parse", "--abbrev-ref", "HEAD")

    @property
    def on_master(self) -> bool:
        return self.git_branch == "master"

    def get_ticket_url(self):
        try:
            with io.open(
                self.get_metadata_path("issue_url"), "r", encoding="utf-8"
            ) as in_:
                return in_.read().strip()
        except (IOError, OSError):
            return None

    def get_metadata_path(self, *args) -> str:
        return os.path.join(self.metadata_dir, *args)

    def get_remote_file_metadata(self, shadow=True):
        remote_files = self.get_path(".jirafs/remote_files.json", shadow=shadow)
        try:
            with io.open(remote_files, "r", encoding="utf-8") as _in:
                data = json.loads(_in.read())
        except IOError:
            data = {}

        return data

    def set_remote_file_metadata(self, data, shadow=True):
        remote_files = self.get_path(".jirafs/remote_files.json", shadow=shadow)
        with io.open(remote_files, "w", encoding="utf-8") as out:
            out.write(json.dumps(data, indent=4, sort_keys=True, ensure_ascii=False,))

    def get_local_path(self, *args):
        return os.path.join(self.path, *args)

    def get_shadow_path(self, *args):
        return os.path.join(self.get_metadata_path("shadow"), *args)

    def get_path(self, *args, **kwargs):
        shadow = kwargs.get("shadow", False)

        if shadow:
            return self.get_shadow_path(*args)
        return self.get_local_path(*args)

    @property
    def version(self):
        try:
            with io.open(
                self.get_metadata_path("version"), "r", encoding="utf-8"
            ) as _in:
                return int(_in.read().strip())
        except IOError:
            return 1

    @property
    def log_path(self):
        return self.get_metadata_path(constants.TICKET_OPERATION_LOG)

    @classmethod
    def initialize_ticket_folder(cls, ticket_url, path, jira):
        path = os.path.realpath(path)

        metadata_path = os.path.join(path, constants.METADATA_DIR,)
        os.mkdir(metadata_path)

        with io.open(
            os.path.join(metadata_path, "issue_url"), "w", encoding="utf-8"
        ) as out:
            out.write(ticket_url)

        # Create bare git repository so we can easily detect changes.
        subprocess.check_call(
            ("git", "--bare", "init", os.path.join(metadata_path, "git",)),
            stdout=subprocess.PIPE,
        )
        subprocess.check_call(
            (
                "git",
                "config",
                "--file=%s" % os.path.join(metadata_path, "git", "config"),
                "core.excludesfile",
                constants.GIT_IGNORE_FILE,
            )
        )

        excludes_path = os.path.join(metadata_path, "git", "info", "exclude")
        with io.open(excludes_path, "w", encoding="utf-8") as gitignore:
            gitignore.write(
                "\n".join(
                    [
                        "%s/git" % constants.METADATA_DIR,
                        "%s/shadow" % constants.METADATA_DIR,
                        "%s/operation.log" % constants.METADATA_DIR,
                    ]
                )
            )

        instance = cls(path, jira, migrate=False)
        instance.log(
            "Ticket folder for issue %s created at %s",
            (instance.ticket_number, instance.path,),
        )
        instance.run_git_command("add", "-A")
        instance.run_git_command("commit", "--allow-empty", "-m", "Initialized")
        instance.run_migrations(init=True)

        comment_path = instance.get_local_path(constants.TICKET_NEW_COMMENT)
        with io.open(comment_path, "w", encoding="utf-8") as out:
            out.write("")

        return instance

    def run_git_command(self, command, *args, **kwargs):
        failure_ok = kwargs.get("failure_ok", False)
        shadow = kwargs.get("shadow", False)
        binary = kwargs.get("binary", False)
        stdin = kwargs.get("stdin", "")

        args = list(args)

        if not shadow:
            work_tree = (self.path,)
            git_dir = self.get_metadata_path("git")
            cwd = self.path
        else:
            work_tree = self.get_metadata_path("shadow")
            git_dir = self.get_metadata_path("shadow/.git")
            cwd = self.get_metadata_path("shadow")

        cmd = [
            "git",
            "--work-tree=%s" % work_tree,
            "--git-dir=%s" % git_dir,
        ]
        cmd.append(command)
        if command == "commit":
            args.append("--author='%s'" % constants.GIT_AUTHOR)
        cmd.extend(args)

        self.log("Executing git command `%s`", (" ".join(cmd),), logging.DEBUG)

        handle = subprocess.Popen(
            cmd,
            cwd=cwd,
            stdin=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            stdout=subprocess.PIPE,
        )
        result, _ = handle.communicate(stdin)

        if handle.returncode != 0 and not failure_ok:
            command = " ".join(cmd)
            raise exceptions.GitCommandError(
                "Error running command `%s`" % command,
                returncode=handle.returncode,
                stdout=result,
                cmd=command,
            )
        if not binary:
            return result.decode("utf-8").strip()
        return result

    def get_local_file_at_revision(self, path, revision, failure_ok=True, binary=False):
        return self.run_git_command(
            "show", "%s:%s" % (revision, path,), failure_ok=failure_ok, binary=binary,
        )

    def get_ignore_globs(self, which=constants.LOCAL_ONLY_FILE):
        all_globs = [
            constants.TICKET_DETAILS,
            constants.TICKET_COMMENTS,
            constants.TICKET_NEW_COMMENT,
            constants.TICKET_LINKS,
        ]
        for field in constants.FILE_FIELDS:
            all_globs.append(
                constants.TICKET_FILE_FIELD_TEMPLATE.format(field_name=field)
            )

        def get_globs_from_file(input_file):
            globs = []
            for line in input_file.readlines():
                if line.startswith("#") or not line.strip():
                    continue
                globs.append(line.strip())
            return globs

        try:
            with io.open(
                self.get_local_path(which), "r", encoding="utf-8"
            ) as local_ign:
                all_globs.extend(get_globs_from_file(local_ign))
        except IOError:
            pass

        try:
            with io.open(
                os.path.expanduser("~/%s" % which), "r", encoding="utf-8"
            ) as global_ignores:
                all_globs.extend(get_globs_from_file(global_ignores))
        except IOError:
            pass

        return all_globs

    def file_matches_globs(self, filename, ignore_globs):
        for glob in ignore_globs:
            if fnmatch.fnmatch(filename, glob):
                return True
        return False

    def get_ready_changes(self):
        ready = {
            "fields": (self.get_fields("HEAD") - self.get_fields(self.git_merge_base)),
            "links": (self.get_links("HEAD") - self.get_links(self.git_merge_base)),
            "new_comment": self.get_new_comment(ready=True),
        }

        current_hash = self.run_git_command("rev-parse", "master")
        committed_files = set(
            self.run_git_command("ls-tree", "--name-only", "-r", current_hash).split(
                "\n"
            )
        )
        merge_base_files = set(
            self.run_git_command(
                "ls-tree", "--name-only", "-r", self.git_merge_base
            ).split("\n")
        )
        ready["deleted"] = self.filter_ignored_files(
            list(merge_base_files - committed_files),
            constants.LOCAL_ONLY_FILE,
            constants.GIT_IGNORE_FILE,
            constants.GIT_EXCLUDE_FILE,
            allow_nonfile=True,
        )

        changed_files = self.filter_ignored_files(
            self.run_git_command(
                "diff", "--name-only", "%s..master" % self.git_merge_base,
            ).split("\n"),
            constants.LOCAL_ONLY_FILE,
        )
        ready["files"] = [
            filename for filename in changed_files if filename not in ready["deleted"]
        ]

        return ready

    def get_uncommitted_changes(self):
        uncommitted = {
            "fields": self.get_fields() - self.get_fields("HEAD"),
            "new_comment": self.get_new_comment(ready=False),
            "links": self.get_links() - self.get_links("HEAD"),
        }

        new_files = self.run_git_command("ls-files", "-o", failure_ok=True).split("\n")
        modified_files = self.run_git_command("ls-files", "-m", failure_ok=True).split(
            "\n"
        )
        deleted_files = self.run_git_command("ls-files", "-d", failure_ok=True).split(
            "\n"
        )
        uncommitted["files"] = self.filter_ignored_files(
            [filename for filename in new_files + modified_files if filename],
            constants.LOCAL_ONLY_FILE,
            constants.GIT_IGNORE_FILE,
            constants.GIT_EXCLUDE_FILE,
        )
        uncommitted["deleted"] = self.filter_ignored_files(
            [filename for filename in deleted_files if filename],
            constants.LOCAL_ONLY_FILE,
            constants.GIT_IGNORE_FILE,
            constants.GIT_EXCLUDE_FILE,
            allow_nonfile=True,  # They're deleted, after all
        )
        return uncommitted

    def get_local_uncommitted_changes(self):
        new_files = self.run_git_command("ls-files", "-o", failure_ok=True).split("\n")
        modified_files = self.run_git_command("ls-files", "-m", failure_ok=True).split(
            "\n"
        )

        committable = self.filter_ignored_files(
            [filename for filename in new_files + modified_files if filename],
            constants.LOCAL_ONLY_FILE,
        )
        uncommitted = self.filter_ignored_files(
            [
                filename
                for filename in modified_files + new_files
                if filename not in committable
            ],
            constants.GIT_IGNORE_FILE,
            constants.GIT_EXCLUDE_FILE,
        )

        return {"files": uncommitted}

    def get_remotely_changed(self):
        metadata = self.get_remote_file_metadata(shadow=True)

        assets = []
        attachments = self.filter_ignored_files(
            getattr(self.issue.fields, "attachment", []), constants.REMOTE_IGNORE_FILE
        )
        for attachment in attachments:
            changed = metadata.get(attachment.filename) != attachment.created
            if changed:
                assets.append(attachment.filename)

        return assets

    def filter_ignored_files(self, files, *which, allow_nonfile=False):
        if len(which) < 1:
            which = [constants.LOCAL_ONLY_FILE]
        if not isinstance(which, (list, tuple)):
            which = [which]

        for list_path in which:
            ignore_globs = self.get_ignore_globs(list_path)

            assets = []
            for fileish in files:
                # Get the actual filename; this is a little gross -- apologies.
                filename = fileish
                attachment = False
                if not isinstance(fileish, str):
                    filename = fileish.filename
                    attachment = True

                if self.file_matches_globs(filename, ignore_globs):
                    continue
                if (
                    not attachment
                    and not allow_nonfile
                    and not os.path.isfile(os.path.join(self.path, filename))
                ):
                    continue
                if filename.startswith("."):
                    continue
                assets.append(fileish)
            files = assets

        return assets

    def get_macro_plugins(self):
        if not hasattr(self, "_macro_plugins"):
            config = self.get_config()
            plugins = []

            if not config.has_section(constants.CONFIG_PLUGINS):
                return plugins

            installed_plugins = utils.get_installed_plugins(MacroPlugin)

            for entrypoint_name, status in config.items(constants.CONFIG_PLUGINS):
                if not utils.convert_to_boolean(status):
                    # This plugin is not turned on.
                    continue
                if entrypoint_name not in installed_plugins:
                    # This plugin is not installed.
                    self.log(
                        "Macro plugin '%s' is not available; "
                        "this is probably because this plugin is not a "
                        "macro.",
                        (entrypoint_name,),
                        level=logging.DEBUG,
                    )
                    continue

                plugin = installed_plugins[entrypoint_name](self, entrypoint_name)

                try:
                    plugin.validate()
                except PluginValidationError as e:
                    self.log(
                        "Plugin '%s' did not pass validation; " "not loading: %s.",
                        (entrypoint_name, e,),
                    )

                plugins.append(plugin)

            self._macro_plugins = plugins

        return self._macro_plugins

    def process_macros_for_all_fields(self):
        # Now let each plugin run its cleanup if necessary
        for plugin in self.get_macro_plugins():
            try:
                plugin.cleanup_pre_process()
            except NotImplementedError:
                pass

        # This is run just in case these macros are writing
        # files as part of their operation, and we need to have
        # those files written in advance of certain operations
        # like listing changes or committing
        fields = self.get_fields()
        for field_name in fields:
            fields.get_transformed(field_name)

        self.get_new_comment()
        with open(self.get_path(constants.TICKET_COMMENTS), "r") as inf:
            self.process_macros(inf.read())

        # Now let each plugin run its cleanup if necessary
        for plugin in self.get_macro_plugins():
            try:
                plugin.cleanup_post_process()
            except NotImplementedError:
                pass

    def process_macros(self, data, path=None):
        macro_plugins = self.get_macro_plugins()

        for plugin in macro_plugins:
            try:
                if isinstance(data, str):
                    data = plugin.process_text_data(data, path)
                else:
                    continue
            except MacroError as e:
                # Annotate the MacroError with information about what
                # macro caused the error
                e.macro_name = plugin.entrypoint_name
                raise e from e

        unprocessed = re.compile(r"(<jirafs:.*>)", re.MULTILINE | re.DOTALL).findall(
            data
        )
        if unprocessed:
            raise exceptions.UnknownMacroError(unprocessed)

        return data

    def process_macro_reversals(self, data):
        macro_plugins = self.get_macro_plugins()

        for plugin in macro_plugins:
            try:
                if isinstance(data, str):
                    data = plugin.process_text_data_reversal(data)
                else:
                    continue
            except MacroError as e:
                # Annotate the MacroError with information about what
                # macro caused the error
                e.macro_name = plugin.entrypoint_name
                raise e from e

        return data

    def get_links(self, revision=None, path=None):
        kwargs = {}
        if not revision:
            kwargs["path"] = path if path else self.path
        else:
            kwargs["revision"] = revision
        return JiraLinkManager.create(self, **kwargs)

    def get_fields(self, revision=None, path=None):
        kwargs = {}
        if not revision:
            kwargs["path"] = path if path else self.path
        else:
            kwargs["revision"] = revision
        return JiraFieldManager.create(self, **kwargs)

    def get_new_comment(self, clear=False, staged=False, ready=False):
        try:
            with io.open(
                self.get_local_path(constants.TICKET_NEW_COMMENT),
                "r+",
                encoding="utf-8",
            ) as c:
                local_contents = c.read().strip()
            if ready:
                contents = self.get_local_file_at_revision(
                    constants.TICKET_NEW_COMMENT, "HEAD"
                )
                if contents:
                    contents = contents.strip()
                else:
                    contents = ""
            else:
                contents = local_contents

            if not ready and contents == self.get_new_comment(ready=True):
                contents = ""

            if contents == local_contents and clear:
                with io.open(
                    self.get_local_path(constants.TICKET_NEW_COMMENT),
                    "r+",
                    encoding="utf-8",
                ) as c:
                    c.truncate()
        except IOError:
            contents = ""

        # Apply macro plugins
        return self.process_macros(contents)

    def get_field_value_by_dotpath(self, field_name, raw=False, **kwargs):
        fields = self.get_fields()

        key_dotpath = None
        if "." in field_name:
            field_name, key_dotpath = field_name.split(".", 1)

        if field_name not in fields:
            raise exceptions.JirafsError("Field '%s' does not exist." % field_name)

        if raw:
            data = fields[field_name]
        else:
            data = fields.get_transformed(field_name)

        if key_dotpath:
            try:
                for component in key_dotpath.split("."):
                    if not isinstance(data, dict):
                        raise exceptions.JirafsError(
                            "Key '%s' (of dotpath '%s') is not an object "
                            "in field '%s'." % (component, key_dotpath, field_name,)
                        )
                    elif component not in data:
                        if "default" in kwargs:
                            data = kwargs["default"]
                        else:
                            raise exceptions.JirafsError(
                                f"Keypath {key_dotpath} does not exist"
                            )
                        break
                    else:
                        data = data[component]
            except (ValueError, TypeError):
                raise exceptions.JirafsError(
                    "Field '%s' could not be parsed as JSON for retrieving "
                    "dotpath '%s'." % (field_name, key_dotpath,)
                )

        return data

    def is_up_to_date(self):
        jira_commit = self.run_git_command("rev-parse", "jira")
        master_commit = self.run_git_command("rev-parse", "master")

        try:
            self.run_git_command(
                "merge-base", "--is-ancestor", jira_commit, master_commit,
            )
        except exceptions.GitCommandError:
            return False
        return True

    def status(self):
        self.process_macros_for_all_fields()

        return {
            "ready": self.get_ready_changes(),
            "local_uncommitted": self.get_local_uncommitted_changes(),
            "uncommitted": self.get_uncommitted_changes(),
            "up_to_date": self.is_up_to_date(),
        }

    def run_migrations(self, init=False):
        loglevel = logging.INFO
        if init:
            loglevel = logging.DEBUG
        else:
            if not os.path.exists(self.get_metadata_path("git")):
                raise exceptions.JirafsError(
                    "{path} is not a valid ticket folder!".format(path=self.path)
                )

            if self.version < constants.CURRENT_REPO_VERSION:
                print(
                    "Your ticket folder at {path} is out-of-date "
                    "and is being automatically updated.".format(path=self.path)
                )
        while self.version < constants.CURRENT_REPO_VERSION:
            migrator = getattr(
                migrations, "migration_%s" % str(self.version + 1).zfill(4)
            )
            self.migrate(migrator, loglevel=loglevel, init=init)

    def migrate(self, migrator, loglevel=logging.INFO, init=False):
        with utils.stash_local_changes(self):
            self.log("%s: Migration started", (migrator.__name__,), loglevel)
            migrator(self, init=init)
            self.log("%s: Migration finished", (migrator.__name__,), loglevel)

    def build_ignore_files(self):
        metadata_excludes = [
            "git",
            "shadow",
            "operation.log",
            "subtasks",
            "temp-generated",
        ]
        with codecs.open(
            self.get_local_path(constants.GIT_EXCLUDE_FILE), "w", "utf-8"
        ) as out:
            for line in metadata_excludes:
                out.write("%s/%s\n" % (constants.METADATA_DIR, line,))
            subtask_list_path = self.get_metadata_path("subtasks")
            if os.path.exists(subtask_list_path):
                with open(subtask_list_path, "r") as in_:
                    for line in in_:
                        out.write("%s/*\n" % line.strip())

        with codecs.open(
            self.get_metadata_path("combined_ignore"), "w", "utf-8"
        ) as out:
            try:
                out.write("# ~/%s\n" % constants.GIT_IGNORE_FILE_PARTIAL)
                with codecs.open(
                    os.path.expanduser("~/%s" % constants.GIT_IGNORE_FILE_PARTIAL),
                    "r",
                    "utf-8",
                ) as in_:
                    for line in in_:
                        out.write("%s\n" % line.strip())
            except Exception:
                pass

            try:
                out.write("# %s\n" % (self.get_path(constants.GIT_IGNORE_FILE_PARTIAL)))
                with codecs.open(
                    self.get_path(constants.GIT_IGNORE_FILE_PARTIAL), "r", "utf-8"
                ) as in_:
                    for line in in_:
                        out.write("%s\n" % line.strip())
            except Exception:
                pass

    def log(self, message, args=None, level=logging.INFO):
        if args is None:
            args = []

        self.logger.log(level, message, *args)

    def get_log(self):
        with io.open(self.log_path, "r", encoding="utf-8") as log_file:
            return log_file.read()



import io
import json
import os
import textwrap

from jirafs import constants, utils
from jirafs.plugin import CommandPlugin


class Command(CommandPlugin):
    """ Fetch remote changes """

    TRY_SUBFOLDERS = True
    RUN_FOR_SUBTASKS = False
    MIN_VERSION = "2.0.0"
    MAX_VERSION = "3.0.0"

    def get_field_map(self, folder):
        fields = {}
        for field in folder.jira.fields():
            fields[field["id"]] = field["name"]

        return fields

    def main(self, folder, **kwargs):
        folder.clear_cache()

        file_meta = folder.get_remote_file_metadata(shadow=True)
        original_hash = folder.run_git_command("rev-parse", "jira")

        for filename in folder.get_remotely_changed():
            for attachment in folder.issue.fields.attachment:
                if attachment.filename == filename:
                    folder.log(
                        'Download file "%s"', (attachment.filename,),
                    )
                    content = io.BytesIO(attachment.get())
                    save_path = folder.get_shadow_path(filename)
                    with open(save_path, "wb") as save_file:
                        content.seek(0)
                        save_file.write(content.read())
                        file_meta[filename] = attachment.created

        folder.set_remote_file_metadata(file_meta, shadow=True)

        field_map = self.get_field_map(folder)
        detail_path = folder.get_shadow_path(constants.TICKET_DETAILS)
        with io.open(detail_path, "w", encoding="utf-8") as dets:
            for field in sorted(folder.issue.raw["fields"].keys()):
                value = folder.issue.raw["fields"][field]
                if isinstance(value, str):
                    value = value.replace("\r\n", "\n").strip()
                elif value is None:
                    value = ""
                elif field in constants.NO_DETAIL_FIELDS:
                    continue

                if not isinstance(value, str):
                    value = json.dumps(
                        value, sort_keys=True, indent=4, ensure_ascii=False
                    )
                else:
                    value = folder.process_macro_reversals(value)

                if field in constants.FILE_FIELDS:
                    # Write specific fields to their own files without
                    # significant alteration

                    file_field_path = folder.get_shadow_path(
                        constants.TICKET_FILE_FIELD_TEMPLATE
                    ).format(field_name=field)
                    with io.open(
                        file_field_path, "w", encoding="utf-8"
                    ) as file_field_file:
                        file_field_file.write(str(value))
                        file_field_file.write("\n")  # For unix' sake
                else:
                    # Normal fields, though, just go into the standard
                    # fields file.
                    if value is None:
                        continue
                    elif field in constants.NO_DETAIL_FIELDS:
                        continue

                    human_readable = field_map.get(field)
                    dets.write("* %s (%s):\n" % (human_readable, field))
                    for line in value.replace("\r\n", "\n").split("\n"):
                        dets.write("    %s\n" % line)

        links_path = folder.get_shadow_path(constants.TICKET_LINKS)
        with io.open(links_path, "w", encoding="utf-8") as links_handle:
            # Write issue links
            for link in folder.issue.fields.issuelinks:
                category = "outward"
                if "inwardIssue" in link.raw:
                    category = "inward"

                links_handle.write(
                    "* {status}: {key}\n".format(
                        status=getattr(link.type, category).title(),
                        key=getattr(link, "%sIssue" % category).key,
                    )
                )

            # Write remote links
            for link in folder.jira.remote_links(folder.issue):
                if link.object.title:
                    links_handle.write(
                        "* {title}: {url}\n".format(
                            title=link.object.title, url=link.object.url
                        )
                    )
                else:
                    links_handle.write(
                        "* {url}\n".format(title=link.object.title, url=link.object.url)
                    )

        comments_filename = folder.get_shadow_path(constants.TICKET_COMMENTS)
        with io.open(comments_filename, "w", encoding="utf-8") as comm:
            for comment in folder.issue.fields.comment.comments:
                comm.write(
                    "h3. At %s, %s wrote:\n\n" % (comment.created, comment.author)
                )
                final_lines = []
                lines = folder.process_macro_reversals(
                    comment.body.replace("\r\n", "\n")
                ).split("\n")
                for line in lines:
                    if not line:
                        final_lines.append("")
                    else:
                        final_lines.extend(
                            textwrap.wrap(
                                line,
                                width=70,
                                expand_tabs=False,
                                replace_whitespace=False,
                                break_long_words=False,
                            )
                        )
                for line in final_lines:
                    comm.write("%s\n" % line)
                comm.write("\n")

        folder.store_cached_issue()

        # Clone subtasks
        subtasks = folder.issue.fields.subtasks
        if len(subtasks) > 0:
            commands = utils.get_installed_commands()
            jira = utils.lazy_get_jira()
            with open(folder.get_metadata_path("subtasks"), "w") as out:
                for issue in subtasks:
                    out.write("%s\n" % issue.key)
                    issue_path = folder.get_path(issue.key)
                    if not os.path.exists(issue_path):
                        command_name = "clone"
                        args = [issue.permalink(), issue_path]
                        path = folder.path
                        commands[command_name].execute_command(
                            args, jira=jira, path=path, command_name=command_name
                        )
        folder.build_ignore_files()

        folder.run_git_command("add", "-A", shadow=True)
        folder.run_git_command(
            "commit", "-m", "Fetched remote changes", failure_ok=True, shadow=True
        )

        folder.run_git_command("push", "origin", "jira", shadow=True)
        final_hash = folder.run_git_command("rev-parse", "jira")
        if original_hash != final_hash:
            folder.log("Updated 'jira' to %s" % final_hash)

        return utils.PostStatusResponse(original_hash == final_hash, final_hash)



import io
import json
import os
import shutil
import tempfile

import mock
from mock import patch
import six

from jirafs import exceptions
from jirafs.utils import run_command_method_with_kwargs
from jirafs.jirafieldmanager import JiraFieldManager

from .base import BaseTestCase


class TestTicketFolder(BaseTestCase):
    def setUp(self):
        self.arbitrary_ticket_number = "ALPHA-123"
        self.root_folder = tempfile.mkdtemp()
        self.mock_jira = mock.MagicMock()
        self.mock_jira.issue.return_value = self.rehydrate_issue("basic.issue.json")
        self.mock_get_jira = lambda _, config=None: self.mock_jira

        with patch(
            "jirafs.ticketfolder.TicketFolder.get_remotely_changed"
        ) as get_remotely_changed:
            get_remotely_changed.return_value = []
            self.ticketfolder = run_command_method_with_kwargs(
                "clone",
                url="http://arbitrary.com/browse/ALPHA-123",
                jira=self.mock_get_jira,
                path=os.path.join(self.root_folder, self.arbitrary_ticket_number,),
            )

    def test_cloned_issue_successfully(self):
        paths = [
            "comments.read_only.jira",
            "fields.jira",
            "description.jira",
            "new_comment.jira",
            ".jirafs/issue.json",
            ".jirafs/remote_files.json",
            ".jirafs/version",
        ]
        for path in paths:
            self.assertTrue(
                os.path.isfile(self.ticketfolder.get_local_path(path)),
                "%s does not exist" % path,
            )
            # Ensure that the file is tracked in git
            try:
                self.ticketfolder.run_git_command("ls-files", path, "--error-unmatch")
            except exceptions.GitCommandError:
                self.fail("File %s is not tracked in the git repository.")

    def test_get_fields(self):
        actual_result = self.ticketfolder.get_fields()

        expected_result = json.loads(self.get_asset_contents("basic.status.json"))

        self.assertEquals(actual_result, expected_result)

    def test_fetch(self):
        self.ticketfolder._issue = self.rehydrate_issue("test_fetch/fetched.json")
        with patch.object(self.ticketfolder, "clear_cache") as clear_cache:
            run_command_method_with_kwargs(
                "fetch", folder=self.ticketfolder,
            )
            self.assertTrue(clear_cache.called)

        expected_result = JiraFieldManager(
            self.get_asset_contents("test_fetch/fetched.jira")
        )
        with io.open(
            self.ticketfolder.get_shadow_path("fields.jira"), encoding="utf-8"
        ) as _in:
            actual_result = JiraFieldManager(_in.read())

        self.assertEqual(actual_result, expected_result)

    def test_push(self):
        changed_field = u"description"
        changed_value = u"Something Else"

        status = self.get_empty_status()
        status["ready"]["fields"][changed_field] = (
            u"Something",
            changed_value,
            changed_value,
        )
        with patch("jirafs.commands.pull.Command.main") as pull:
            with patch.object(self.ticketfolder, "status") as status_method:
                status_method.return_value = status
                with patch.object(self.ticketfolder.issue, "update") as out:
                    run_command_method_with_kwargs("push", folder=self.ticketfolder)
                    self.assertTrue(pull.called)
                    out.assert_called_with(**{"description": changed_value})

    def test_push_rejected_if_updated(self):
        src_path = self.get_asset_path("test_fetch/fetched.jira")
        dst_path = self.ticketfolder.get_shadow_path("fields.jira")
        shutil.copyfile(
            src_path, dst_path,
        )

        self.ticketfolder.run_git_command("add", "-A", shadow=True)
        self.ticketfolder.run_git_command("commit", "-m", "Changed", shadow=True)
        self.ticketfolder.run_git_command("push", "origin", "jira", shadow=True)
        with self.assertRaises(exceptions.LocalCopyOutOfDate):
            run_command_method_with_kwargs("push", folder=self.ticketfolder)

    def test_status_new_file(self):
        src_path = self.get_asset_path("test_status_local_changes/alpha.svg")
        dst_path = self.ticketfolder.get_local_path("alpha.svg")
        shutil.copyfile(
            src_path, dst_path,
        )

        expected_output = self.get_empty_status()
        expected_output["uncommitted"]["files"] = ["alpha.svg"]
        actual_output = self.ticketfolder.status()

        self.assertEqual(expected_output, actual_output)

    def test_status_local_changes(self):
        src_path = self.get_asset_path("test_fetch/fetched.jira")
        dst_path = self.ticketfolder.get_local_path("fields.jira")
        shutil.copyfile(
            src_path, dst_path,
        )

        expected_output = self.get_empty_status()
        expected_output["uncommitted"]["fields"]["assignee"] = (
            "",
            {
                "active": True,
                "displayName": (
                    u"Coddington, Adam (\u0410\u0440\u0431\u0438\u0442"
                    u"\u0440\u0430\u0440\u0438\u041a\u043e\u0440\u043f"
                    u"-Atlantis)"
                ),
                "name": "acoddington",
            },
            {
                "active": True,
                "displayName": (
                    u"Coddington, Adam (\u0410\u0440\u0431\u0438\u0442"
                    u"\u0440\u0430\u0440\u0438\u041a\u043e\u0440\u043f"
                    u"-Atlantis)"
                ),
                "name": "acoddington",
            },
        )
        actual_output = self.ticketfolder.status()

        for field in expected_output:
            if expected_output[field] != actual_output[field]:
                self.fail(
                    "Field %s does not match; Actual: %s != Expected: %s"
                    % (field, actual_output[field], expected_output[field],)
                )

        self.assertEqual(expected_output, actual_output)

    def test_status_new_comment(self):
        arbitrary_comment = six.text_type("New Comment")

        comment = self.ticketfolder.get_local_path("new_comment.jira")
        with io.open(comment, "w", encoding="utf-8") as out:
            out.write(arbitrary_comment)

        expected_output = self.get_empty_status()
        expected_output["uncommitted"]["new_comment"] = arbitrary_comment
        actual_output = self.ticketfolder.status()

        self.assertEqual(expected_output, actual_output)

    def tearDown(self):
        shutil.rmtree(self.root_folder)



from jirafs.plugin import CommandPlugin

try:
    import ipdb as pdb  # noqa
except ImportError:
    import pdb


class Command(CommandPlugin):
    """ Open a debug console """

    MIN_VERSION = "2.0.0"
    MAX_VERSION = "3.0.0"

    def main(self, folder, **kwargs):
        return pdb.set_trace()



import io
import json
import os
from unittest import TestCase

from jira.resources import Issue


class BaseTestCase(TestCase):
    def get_asset_path(self, filename):
        return os.path.join(os.path.dirname(__file__), "assets", filename)

    def get_asset_contents(self, filename, mode="r"):
        path = self.get_asset_path(filename)

        with io.open(path, mode, encoding="utf-8") as r:
            return r.read()

    def rehydrate_issue(self, filename):
        stored = json.loads(self.get_asset_contents(filename))
        return Issue(stored["options"], None, stored["raw"],)

    def get_empty_status(self):
        return {
            "ready": {
                "files": [],
                "links": {},
                "fields": {},
                "new_comment": "",
                "deleted": [],
            },
            "uncommitted": {
                "files": [],
                "links": {},
                "fields": {},
                "new_comment": "",
                "deleted": [],
            },
            "local_uncommitted": {"files": []},
            "up_to_date": True,
        }



import webbrowser

from jirafs.plugin import CommandPlugin


class Command(CommandPlugin):
    """ Open the current ticketfolder's issue in your web browser """

    MIN_VERSION = "2.0.0"
    MAX_VERSION = "3.0.0"

    def main(self, folder, *args, **kwargs):
        return webbrowser.open(folder.cached_issue.permalink())



from mock import call, patch

from jirafs.utils import run_command_method_with_kwargs

from .base import BaseCommandTestCase


class TestPushCommand(BaseCommandTestCase):
    def setUp(self):
        super(TestPushCommand, self).setUp()

    def test_push_no_changes(self):
        with patch.object(self.ticketfolder.issue, "update") as update:
            with patch("jirafs.commands.pull.Command.main") as pull:
                pull.return_value = True, True
                run_command_method_with_kwargs(
                    "push", folder=self.ticketfolder,
                )

            self.assertEqual(0, len(update.call_args_list))

    def test_push_basic(self):
        status_result = {
            "ready": {
                "files": [],
                "fields": {
                    "somefield": ("one", "two", "two",),
                    "otherfield": ("one", "three", "three",),
                },
                "links": {},
                "deleted": [],
            }
        }

        with patch.object(self.ticketfolder, "status") as status:
            status.return_value = status_result
            with patch.object(self.ticketfolder.issue, "update") as update:
                with patch("jirafs.commands.pull.Command.main") as pull:
                    pull.return_value = True, True
                    run_command_method_with_kwargs(
                        "push", folder=self.ticketfolder,
                    )

                self.assertEqual(1, len(update.call_args_list))
                self.assertEqual(
                    update.call_args,
                    call(
                        **{
                            field_name: value[1]
                            for field_name, value in status_result["ready"][
                                "fields"
                            ].items()
                        }
                    ),
                )



from jirafs.plugin import CommandPlugin


class Command(CommandPlugin):
    """ Print a diff of locally-changed files """

    MIN_VERSION = "2.0.0"
    MAX_VERSION = "3.0.0"

    def main(self, folder, **kwargs):
        result = folder.run_git_command("diff")
        if result:
            result = result.strip()
        print(result)
        return result



from contextlib import closing
import html
import json
import mimetypes
import os
import re
import socket
import time
import traceback
import uuid
import webbrowser
from http.server import ThreadingHTTPServer
from http.server import SimpleHTTPRequestHandler

from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler

import jinja2

from jirafs.plugin import CommandPlugin


class CountingEventHandler(FileSystemEventHandler):
    counter = 0

    def on_modified(self, event):
        self.counter += 1


def get_converted_markup(folder, data):
    url = folder.jira._get_url("../1.0/render")
    response = folder.jira._session.post(
        url,
        headers={"Accept": "text/html, */*; q=0.01"},
        data=json.dumps(
            {
                "issueKey": folder.ticket_number,
                "rendererType": "atlassian-wiki-renderer",
                "unrenderedMarkup": data,
            }
        ),
    )

    return response.text


class IssueRequestHandler(SimpleHTTPRequestHandler):
    folder = None
    field_data = None

    def get_rendered_template(self, template_name, context):
        loader = jinja2.FileSystemLoader(
            searchpath=os.path.join(os.path.dirname(__file__), "templates")
        )
        templates = jinja2.Environment(loader=loader)
        template = templates.get_template(template_name)
        return template.render(context)

    def get_all(self):
        lines = []

        lines.append(
            f"h1. {self.folder.issue.id}: {self.get_field_data('summary')}\n\n"
        )
        lines.append(f"h2. Description\n\n")
        lines.append(self.get_field_data("description"))
        lines.append("\n")
        lines.append(f"h2. New Comment\n\n")
        lines.append(self.get_field_data("new_comment"))

        return "\n".join(lines)

    def get_comments(self):
        lines = []

        for comment in self.folder.issue.fields.comment.comments:
            lines.append("h3. At %s, %s wrote:\n" % (comment.created, comment.author,))
            lines.append(comment.body.replace("\r\n", "\n"))

        return "\n".join(lines)

    def get_field_data(self, dotpath):
        special_fields = {
            "": self.get_all,
            "new_comment": self.folder.get_new_comment,
            "comments": self.get_comments,
        }
        if dotpath in special_fields:
            data = special_fields[dotpath]()
        else:
            data = self.folder.get_field_value_by_dotpath(dotpath)

        return data

    def get_local_file_escaped_field_data(self, dotpath):
        data = self.get_field_data(dotpath)

        if not data:
            return {}, ""

        local_files = os.listdir(self.folder.path)

        to_replace = []
        finders = [
            re.compile(r"(!(?P<filename>[^!]+)!)"),
            re.compile(r"(\[\^(?P<filename>[^\]]+)\])"),
        ]
        for finder in finders:
            for match in finder.finditer(data):
                filename = match.groupdict()["filename"]
                if "|" in filename:
                    filename = filename.split("|", 1)[0]

                if filename in local_files:
                    to_replace.append(
                        (filename, match.group(0), match.start(0), match.end(0),)
                    )

        placeholders = {}
        for filename, full, start, end in reversed(to_replace):
            id = uuid.uuid4()
            placeholder = f"JIRAFS-PLACEHOLDER:{id}"
            placeholders[placeholder] = (filename, full)
            data = "".join([data[:start], placeholder, data[end:]])

        return placeholders, data

    def replace_placeholders(self, placeholders, data):
        for placeholder, (filename, full) in placeholders.items():
            if full.startswith("!"):
                data = data.replace(placeholder, f'<img src="files/{filename}" />')
            elif full.startswith("[^"):
                data = data.replace(
                    placeholder, f'<a href="files/{filename}">{filename}</a>'
                )

        return data

    def serve_preview_content(self, dotpath):
        content_type = "text/html"
        placeholders, data = self.get_local_file_escaped_field_data(dotpath)
        if isinstance(data, str):
            response = self.get_rendered_template(
                "base.html",
                {
                    "content": self.replace_placeholders(
                        placeholders, get_converted_markup(self.folder, data)
                    )
                },
            )
        else:
            response = json.dumps(data)
            content_type = "application/json"
        self.send_response(200)
        self.send_header("Content-type", content_type)
        self.send_header("Content-length", len(response))
        self.end_headers()
        self.wfile.write(response.encode("utf-8"))

    def serve_file(self, path):
        print(path)

        if path not in os.listdir(self.folder.path):
            self.send_response(404)
            self.send_header("Content-length", 0)
            self.end_headers()
            return

        with open(os.path.join(self.folder.path, path), "rb") as inf:
            self.send_response(200)
            inf.seek(0, 2)
            content_type = mimetypes.guess_type(path)[0] or "application/octet-stream"
            self.send_header("Content-type", content_type)
            self.send_header("Content-length", inf.tell())
            self.end_headers()
            inf.seek(0)
            self.wfile.write(inf.read())

    def send_eventsource_message(self, message):
        self.wfile.write(str(len(message.encode("utf-8")) + 2).encode("utf-8"))
        self.wfile.write("\r\n".encode("utf-8"))
        self.wfile.write(message.encode("utf-8"))
        self.wfile.write("\r\n".encode("utf-8"))

    def serve_eventsource(self):
        event_handler = CountingEventHandler()
        observer = Observer()
        observer.schedule(event_handler, path=self.folder.path, recursive=True)
        observer.start()

        self.send_response(200)
        self.send_header("Transfer-encoding", "chunked")
        self.send_header("Content-Type", "text/event-stream")
        self.end_headers()

        while True:
            self.send_eventsource_message(
                "event: counter\r\ndata: %s\r\n" % event_handler.counter
            )
            time.sleep(0.5)

    def do_GET(self):
        try:
            if self.path.startswith("/files/"):
                self.serve_file(self.path[7:])
            elif self.path == "/eventsource/":
                self.serve_eventsource()
            else:
                self.serve_preview_content(self.path[1:].replace("/", "."))
        except Exception:
            self.send_response(500)
            response = self.get_rendered_template(
                "traceback.html", {"content": html.escape(traceback.format_exc())}
            )
            self.send_header("Content-type", "text/html")
            self.send_header("Content-length", len(response))
            self.end_headers()
            self.wfile.write(response.encode("utf-8"))


class Command(CommandPlugin):
    """ Preview your Jira wiki markup """

    MIN_VERSION = "2.0.0"
    MAX_VERSION = "3.0.0"

    def add_arguments(self, parser):
        parser.add_argument(
            "--port",
            "-p",
            help=(
                "Start a webserver on this port; defaults to asking "
                "the operating system for any available port."
            ),
            type=int,
            default=0,
        )
        parser.add_argument(
            "--no-browser",
            "-n",
            action="store_true",
            default=False,
            help=("Do not open a webbrowser to the created webserver."),
        )
        parser.add_argument("field_name", nargs="?")

    def handle(self, args, folder, **kwargs):
        return self.cmd(
            folder,
            args.field_name or "",
            port=args.port,
            open_browser=not args.no_browser,
        )

    def main(self, folder, field_name, port=0, open_browser=True, **kwargs):
        with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
            s.bind(("", port))
            s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            port = s.getsockname()[1]
            path = field_name.replace(".", "/")

            IssueRequestHandler.folder = folder
            IssueRequestHandler.get_converted_markup = get_converted_markup

            server = ThreadingHTTPServer(("", port), IssueRequestHandler)
            print(f"Serving from http://127.0.0.1:{port}")
            print("Press <Ctrl+C> to Exit")

            if open_browser:
                webbrowser.open(f"http://127.0.0.1:{port}/{path}")

            while True:
                server.handle_request()

            server.shutdown()



import os
from setuptools import setup
from setuptools.command.test import test as TestCommand
import sys

from jirafs import __version__ as version_string


requirements_path = os.path.join(os.path.dirname(__file__), "requirements.txt",)
requirements = []
with open(requirements_path, "r") as in_:
    requirements = [
        req.strip()
        for req in in_.readlines()
        if not req.startswith("-") and not req.startswith("#")
    ]


class Tox(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import tox  # noqa

        errno = tox.cmdline(self.test_args)
        sys.exit(errno)


setup(
    name="jirafs",
    version=version_string,
    url="https://github.com/coddingtonbear/jirafs",
    description=("Edit Jira issues as text files locally."),
    author="Adam Coddington",
    author_email="me@adamcoddington.net",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
    ],
    install_requires=requirements,
    tests_require=["tox", "behave>=1.2.5"],
    cmdclass={"test": Tox},
    packages=["jirafs"],
    include_package_data=True,
    entry_points={
        "console_scripts": ["jirafs = jirafs.cmdline:main"],
        "jirafs_commands": [
            "fetch = jirafs.commands.fetch:Command",
            "pull = jirafs.commands.pull:Command",
            "merge = jirafs.commands.merge:Command",
            "commit = jirafs.commands.commit:Command",
            "push = jirafs.commands.push:Command",
            "git = jirafs.commands.git:Command",
            "log = jirafs.commands.log:Command",
            "debug = jirafs.commands.debug:Command",
            "plugins = jirafs.commands.plugins:Command",
            "status = jirafs.commands.status:Command",
            "clone = jirafs.commands.clone:Command",
            "open = jirafs.commands.open:Command",
            "diff = jirafs.commands.diff:Command",
            "config = jirafs.commands.config:Command",
            "transition = jirafs.commands.transition:Command",
            "field = jirafs.commands.field:Command",
            "setfield = jirafs.commands.setfield:Command",
            "subtask = jirafs.commands.subtask:Command",
            "assign = jirafs.commands.assign:Command",
            "search_users = jirafs.commands.search_users:Command",
            "create = jirafs.commands.create:Command",
            "submit = jirafs.commands.submit:Command",
            "match = jirafs.commands.match:Command",
            "preview = jirafs.commands.preview:Command",
            "version = jirafs.commands.version:Command",
        ],
    },
)
